﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using UdpClientLib;

namespace HomeGenie.Automation.Scripting
{

    /// <summary>
    /// Udp client helper.\n
    /// Class instance accessor: **udpClient**
    /// </summary>
    public class UdpClientHelper
    {
        private UDPClient udpClient;
        private Action<byte[]> dataReceived;
        private Action<string> stringReceived;
        private Action<bool> statusChanged;

        public UdpClientHelper()
        {
            udpClient = new UDPClient();
        }

        /// <summary>
        /// Sets the client as a sender to address:port
        /// </summary>
        /// <returns>udpClientHelper.</returns>
        /// <param name="address">Remote DNS or IP address.</param>
        /// <param name="port">port to send to</param>
        public UdpClientHelper Sender(string address, int port)
        {
            udpClient.Connect(address, port);
            return this;
        }

        /// <summary>
        /// Connect to the remote using the specified port.
        /// </summary>
        /// <param name="port">Port number.</param>
        public bool Receiver(int port)
        {
            udpClient.MessageReceived += udpClient_MessageReceived;
            udpClient.ConnectedStateChanged += udpClient_ConnectedStateChanged;
            //
            return udpClient.Connect(port);
        }

        /// <summary>
        /// Disconnects from the remote host.
        /// </summary>
        public UdpClientHelper Disconnect()
        {
            udpClient.Disconnect();
            udpClient.MessageReceived -= udpClient_MessageReceived;
            udpClient.ConnectedStateChanged -= udpClient_ConnectedStateChanged;
            return this;
        }

        /// <summary>
        /// Sends a string message.
        /// </summary>
        /// <param name="message">Message.</param>
        public void SendMessage(string message)
        {
            if (udpClient.IsConnected)
            {
                byte[] msg = Encoding.UTF8.GetBytes(message);
                udpClient.SendMessage(msg);
            }
        }

        /// <summary>
        /// Sends a raw data message.
        /// </summary>
        /// <param name="message">Message.</param>
        public void SendMessage(byte[] message)
        {
            if (udpClient.IsConnected)
            {
                udpClient.SendMessage(message);
            }
        }

        /// <summary>
        /// Sets the function to call when a new string message is received.
        /// </summary>
        /// <param name="receivedAction">Function or inline delegate.</param>
        public UdpClientHelper OnStringReceived(Action<string> receivedAction)
        {
            stringReceived = receivedAction;
            return this;
        }

        /// <summary>
        /// Sets the function to call when a new raw message is received.
        /// </summary>
        /// <param name="receivedAction">Function or inline delegate.</param>
        public UdpClientHelper OnMessageReceived(Action<byte[]> receivedAction)
        {
            dataReceived = receivedAction;
            return this;
        }

        /// <summary>
        /// Sets the function to call when the status of the connection changes.
        /// </summary>
        /// <param name="statusChangeAction">Function or inline delegate.</param>
        public UdpClientHelper OnStatusChanged(Action<bool> statusChangeAction)
        {
            statusChanged = statusChangeAction;
            return this;
        }

        /// <summary>
        /// Gets a value indicating whether the connection to the service is estabilished.
        /// </summary>
        /// <value><c>true</c> if connected; otherwise, <c>false</c>.</value>
        public bool IsConnected
        {
            get { return udpClient.IsConnected; }
        }


        private void udpClient_MessageReceived(byte[] message)
        {
            if (dataReceived != null)
            {
                dataReceived(message);
            }
            if (stringReceived != null)
            {
                try
                {
                    stringReceived(Encoding.UTF8.GetString(message));
                }
                catch { }
            }
        }

        private void udpClient_ConnectedStateChanged(object sender, ConnectedStateChangedEventArgs statusargs)
        {
            if (statusChanged != null)
            {
                statusChanged(statusargs.Connected);
            }
        }

    }
}
